// Copyright (C) 1999-2000 Id Software, Inc.
//
// cg_effects.c -- these functions generate localentities, usually as a result
// of event processing

/*
 *  Portions Copyright (C) 2000-2001 Tim Angus
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the OSML - Open Source Modification License v1.0 as
 *  described in the file COPYING which is distributed with this source
 *  code.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

#include "cg_local.h"


/*
==================
CG_BubbleTrail

Bullets shot underwater
==================
*/
void CG_BubbleTrail( vec3_t start, vec3_t end, float spacing )
{
  vec3_t    move;
  vec3_t    vec;
  float     len;
  int       i;
  int       sourceContentType, destContentType;
  trace_t   trace;
  vec3_t    s, e;

  if( cg_noProjectileTrail.integer )
    return;

  sourceContentType = trap_CM_PointContents( start, 0 );
  destContentType = trap_CM_PointContents( end, 0 );

  if( ( sourceContentType & CONTENTS_WATER ) && ( destContentType & CONTENTS_WATER ) )
  {
    //do a complete bubble trail if necessary
    VectorCopy( start, s );
    VectorCopy( end, e );
  }
  else if( ( sourceContentType & CONTENTS_WATER ) )
  {
    //bubble trail from water into air
    trap_CM_BoxTrace( &trace, end, start, NULL, NULL, 0, CONTENTS_WATER );
    VectorCopy( start, s );
    VectorCopy( trace.endpos, e );
  }
  else if( ( destContentType & CONTENTS_WATER ) )
  {
    //bubble trail from air into water
    trap_CM_BoxTrace( &trace, start, end, NULL, NULL, 0, CONTENTS_WATER );
    VectorCopy( trace.endpos, s );
    VectorCopy( end, e );
  }
  else
    return;

  VectorCopy( s, move );
  VectorSubtract( e, s, vec );
  len = VectorNormalize( vec );

  // advance a random amount first
  i = rand( ) % (int)spacing;
  VectorMA( move, i, vec, move );

  VectorScale( vec, spacing, vec );

  for( ; i < len; i += spacing )
  {
    localEntity_t *le;
    refEntity_t   *re;

    le = CG_AllocLocalEntity( );
    le->leFlags = LEF_PUFF_DONT_SCALE;
    le->leType = LE_MOVE_SCALE_FADE;
    le->startTime = cg.time;
    le->endTime = cg.time + 1000 + random( ) * 250;
    le->lifeRate = 1.0 / ( le->endTime - le->startTime );

    re = &le->refEntity;
    re->shaderTime = cg.time / 1000.0f;

    re->reType = RT_SPRITE;
    re->rotation = 0;
    re->radius = 3;
    re->customShader = cgs.media.waterBubbleShader;
    re->shaderRGBA[ 0 ] = 0xff;
    re->shaderRGBA[ 1 ] = 0xff;
    re->shaderRGBA[ 2 ] = 0xff;
    re->shaderRGBA[ 3 ] = 0xff;

    le->color[ 3 ] = 1.0;

    le->pos.trType = TR_LINEAR;
    le->pos.trTime = cg.time;
    VectorCopy( move, le->pos.trBase );
    le->pos.trDelta[ 0 ] = crandom( ) * 5;
    le->pos.trDelta[ 1 ] = crandom( ) * 5;
    le->pos.trDelta[ 2 ] = crandom( ) * 5 + 6;

    VectorAdd( move, vec, move );
  }
}

/*
=====================
CG_SmokePuff

Adds a smoke puff or blood trail localEntity.
=====================
*/
localEntity_t *CG_SmokePuff( const vec3_t p, const vec3_t vel,
                             float radius,
                             float r, float g, float b, float a,
                             float duration,
                             int startTime,
                             int fadeInTime,
                             int leFlags,
                             qhandle_t hShader )
{
  static int  seed = 0x92;
  localEntity_t *le;
  refEntity_t   *re;

  le = CG_AllocLocalEntity();
  le->leFlags = leFlags;
  le->radius = radius;

  re = &le->refEntity;
  re->rotation = Q_random( &seed ) * 360;
  re->radius = radius;
  re->shaderTime = startTime / 1000.0f;

  le->leType = LE_MOVE_SCALE_FADE;
  le->startTime = startTime;
  le->fadeInTime = fadeInTime;
  le->endTime = startTime + duration;

  if( fadeInTime > startTime )
    le->lifeRate = 1.0 / ( le->endTime - le->fadeInTime );
  else
    le->lifeRate = 1.0 / ( le->endTime - le->startTime );

  le->color[ 0 ] = r;
  le->color[ 1 ] = g;
  le->color[ 2 ] = b;
  le->color[ 3 ] = a;


  le->pos.trType = TR_LINEAR;
  le->pos.trTime = startTime;
  VectorCopy( vel, le->pos.trDelta );
  VectorCopy( p, le->pos.trBase );

  VectorCopy( p, re->origin );
  re->customShader = hShader;

  // rage pro can't alpha fade, so use a different shader
  if( cgs.glconfig.hardwareType == GLHW_RAGEPRO )
  {
    re->customShader = cgs.media.smokePuffRageProShader;
    re->shaderRGBA[ 0 ] = 0xff;
    re->shaderRGBA[ 1 ] = 0xff;
    re->shaderRGBA[ 2 ] = 0xff;
    re->shaderRGBA[ 3 ] = 0xff;
  }
  else
  {
    re->shaderRGBA[ 0 ] = le->color[ 0 ] * 0xff;
    re->shaderRGBA[ 1 ] = le->color[ 1 ] * 0xff;
    re->shaderRGBA[ 2 ] = le->color[ 2 ] * 0xff;
    re->shaderRGBA[ 3 ] = 0xff;
  }

  re->reType = RT_SPRITE;
  re->radius = le->radius;

  return le;
}

/*
==================
CG_SpawnEffect

Player teleporting in or out
==================
*/
void CG_SpawnEffect( vec3_t org )
{
  localEntity_t *le;
  refEntity_t   *re;

  le = CG_AllocLocalEntity( );
  le->leFlags = 0;
  le->leType = LE_FADE_RGB;
  le->startTime = cg.time;
  le->endTime = cg.time + 500;
  le->lifeRate = 1.0 / ( le->endTime - le->startTime );

  le->color[ 0 ] = le->color[ 1 ] = le->color[ 2 ] = le->color[ 3 ] = 1.0;

  re = &le->refEntity;

  re->reType = RT_MODEL;
  re->shaderTime = cg.time / 1000.0f;

  re->customShader = cgs.media.teleportEffectShader;
  re->hModel = cgs.media.teleportEffectModel;
  AxisClear( re->axis );

  VectorCopy( org, re->origin );
  re->origin[ 2 ] -= 24;
}




/*
====================
CG_MakeExplosion
====================
*/
localEntity_t *CG_MakeExplosion( vec3_t origin, vec3_t dir,
                                 qhandle_t hModel, qhandle_t shader,
                                 int msec, qboolean isSprite )
{
  float         ang;
  localEntity_t *ex;
  int           offset;
  vec3_t        tmpVec, newOrigin;

  if( msec <= 0 )
    CG_Error( "CG_MakeExplosion: msec = %i", msec );

  // skew the time a bit so they aren't all in sync
  offset = rand( ) & 63;

  ex = CG_AllocLocalEntity( );

  if( isSprite )
  {
    ex->leType = LE_SPRITE_EXPLOSION;

    // randomly rotate sprite orientation
    ex->refEntity.rotation = rand( ) % 360;
    VectorScale( dir, 16, tmpVec );
    VectorAdd( tmpVec, origin, newOrigin );
  }
  else
  {
    ex->leType = LE_EXPLOSION;
    VectorCopy( origin, newOrigin );

    // set axis with random rotate
    if( !dir )
      AxisClear( ex->refEntity.axis );
    else
    {
      ang = rand( ) % 360;
      VectorCopy( dir, ex->refEntity.axis[ 0 ] );
      RotateAroundDirection( ex->refEntity.axis, ang );
    }
  }

  ex->startTime = cg.time - offset;
  ex->endTime = ex->startTime + msec;

  // bias the time so all shader effects start correctly
  ex->refEntity.shaderTime = ex->startTime / 1000.0f;

  ex->refEntity.hModel = hModel;
  ex->refEntity.customShader = shader;

  // set origin
  VectorCopy( newOrigin, ex->refEntity.origin );
  VectorCopy( newOrigin, ex->refEntity.oldorigin );

  ex->color[ 0 ] = ex->color[ 1 ] = ex->color[ 2 ] = 1.0;

  return ex;
}


/*
=================
CG_Bleed

This is the spurt of blood when a character gets hit
=================
*/
void CG_Bleed( vec3_t origin, int entityNum )
{
  pTeam_t       team = cgs.clientinfo[ entityNum ].team;
  localEntity_t *ex;

  if( !cg_blood.integer )
    return;

  ex = CG_AllocLocalEntity( );
  ex->leType = LE_EXPLOSION;

  ex->startTime = cg.time;
  ex->endTime = ex->startTime + 300;

  VectorCopy( origin, ex->refEntity.origin );
  ex->refEntity.reType = RT_SPRITE;
  ex->refEntity.rotation = rand( ) % 360;
  ex->refEntity.radius = 8;
  ex->refEntity.shaderTime = cg.time / 1000.0f;

  if( team == PTE_HUMANS )
    ex->refEntity.customShader = cgs.media.bloodExplosionShader;
  else if( team == PTE_ALIENS )
    ex->refEntity.customShader = cgs.media.greenBloodExplosionShader;

  // don't show player's own blood in view
  if( entityNum == cg.snap->ps.clientNum )
    ex->refEntity.renderfx |= RF_THIRD_PERSON;
}



/*
==================
CG_LaunchGib
==================
*/
void CG_LaunchGib( vec3_t origin, vec3_t velocity, qhandle_t hModel )
{
  localEntity_t *le;
  refEntity_t   *re;

  le = CG_AllocLocalEntity( );
  re = &le->refEntity;

  le->leType = LE_FRAGMENT;
  le->startTime = cg.time;
  le->endTime = le->startTime + 5000 + random() * 3000;

  VectorCopy( origin, re->origin );
  AxisCopy( axisDefault, re->axis );
  re->hModel = hModel;

  le->pos.trType = TR_GRAVITY;
  VectorCopy( origin, le->pos.trBase );
  VectorCopy( velocity, le->pos.trDelta );
  le->pos.trTime = cg.time;

  le->bounceFactor = 0.6f;

  le->leBounceSoundType = LEBS_BLOOD;
  le->leMarkType = LEMT_BLOOD;
}

/*
===================
CG_GibPlayer

Generated a bunch of gibs launching out from the bodies location
===================
*/
#define GIB_VELOCITY  250
#define GIB_JUMP    250
void CG_GibPlayer( vec3_t playerOrigin )
{
  vec3_t  origin, velocity;

  if( !cg_blood.integer )
    return;

  VectorCopy( playerOrigin, origin );
  velocity[ 0 ] = crandom( ) * GIB_VELOCITY;
  velocity[ 1 ] = crandom( ) * GIB_VELOCITY;
  velocity[ 2 ] = GIB_JUMP + crandom( ) * GIB_VELOCITY;

  if( rand( ) & 1 )
    CG_LaunchGib( origin, velocity, cgs.media.gibSkull );
  else
    CG_LaunchGib( origin, velocity, cgs.media.gibBrain );

  // allow gibs to be turned off for speed
  if( !cg_gibs.integer )
    return;

  VectorCopy( playerOrigin, origin );
  velocity[ 0 ] = crandom( ) * GIB_VELOCITY;
  velocity[ 1 ] = crandom( ) * GIB_VELOCITY;
  velocity[ 2 ] = GIB_JUMP + crandom( ) * GIB_VELOCITY;
  CG_LaunchGib( origin, velocity, cgs.media.gibAbdomen );

  VectorCopy( playerOrigin, origin );
  velocity[ 0 ] = crandom( ) * GIB_VELOCITY;
  velocity[ 1 ] = crandom( ) * GIB_VELOCITY;
  velocity[ 2 ] = GIB_JUMP + crandom( ) * GIB_VELOCITY;
  CG_LaunchGib( origin, velocity, cgs.media.gibArm );

  VectorCopy( playerOrigin, origin );
  velocity[ 0 ] = crandom( ) * GIB_VELOCITY;
  velocity[ 1 ] = crandom( ) * GIB_VELOCITY;
  velocity[ 2 ] = GIB_JUMP + crandom( ) * GIB_VELOCITY;
  CG_LaunchGib( origin, velocity, cgs.media.gibChest );

  VectorCopy( playerOrigin, origin );
  velocity[ 0 ] = crandom( ) * GIB_VELOCITY;
  velocity[ 1 ] = crandom( ) * GIB_VELOCITY;
  velocity[ 2 ] = GIB_JUMP + crandom( ) * GIB_VELOCITY;
  CG_LaunchGib( origin, velocity, cgs.media.gibFist );

  VectorCopy( playerOrigin, origin );
  velocity[ 0 ] = crandom( ) * GIB_VELOCITY;
  velocity[ 1 ] = crandom( ) * GIB_VELOCITY;
  velocity[ 2 ] = GIB_JUMP + crandom( ) * GIB_VELOCITY;
  CG_LaunchGib( origin, velocity, cgs.media.gibFoot );

  VectorCopy( playerOrigin, origin );
  velocity[ 0 ] = crandom( ) * GIB_VELOCITY;
  velocity[ 1 ] = crandom( ) * GIB_VELOCITY;
  velocity[ 2 ] = GIB_JUMP + crandom( ) * GIB_VELOCITY;
  CG_LaunchGib( origin, velocity, cgs.media.gibForearm );

  VectorCopy( playerOrigin, origin );
  velocity[ 0 ] = crandom( ) * GIB_VELOCITY;
  velocity[ 1 ] = crandom( ) * GIB_VELOCITY;
  velocity[ 2 ] = GIB_JUMP + crandom( ) * GIB_VELOCITY;
  CG_LaunchGib( origin, velocity, cgs.media.gibIntestine );

  VectorCopy( playerOrigin, origin );
  velocity[ 0 ] = crandom( ) * GIB_VELOCITY;
  velocity[ 1 ] = crandom( ) * GIB_VELOCITY;
  velocity[ 2 ] = GIB_JUMP + crandom( ) * GIB_VELOCITY;
  CG_LaunchGib( origin, velocity, cgs.media.gibLeg );

  VectorCopy( playerOrigin, origin );
  velocity[ 0 ] = crandom( ) * GIB_VELOCITY;
  velocity[ 1 ] = crandom( ) * GIB_VELOCITY;
  velocity[ 2 ] = GIB_JUMP + crandom( ) * GIB_VELOCITY;
  CG_LaunchGib( origin, velocity, cgs.media.gibLeg );
}

//id big explosions - must try

/*
==================
CG_LaunchExplode
==================
*/
void CG_LaunchExplode( vec3_t origin, vec3_t velocity, qhandle_t hModel )
{
  localEntity_t *le;
  refEntity_t   *re;

  le = CG_AllocLocalEntity( );
  re = &le->refEntity;

  le->leType = LE_FRAGMENT;
  le->startTime = cg.time;
  le->endTime = le->startTime + 10000 + random( ) * 6000;

  VectorCopy( origin, re->origin );
  AxisCopy( axisDefault, re->axis );
  re->hModel = hModel;

  le->pos.trType = TR_GRAVITY;
  VectorCopy( origin, le->pos.trBase );
  VectorCopy( velocity, le->pos.trDelta );
  le->pos.trTime = cg.time;

  le->bounceFactor = 0.1f;

  le->leBounceSoundType = LEBS_BRASS;
  le->leMarkType = LEMT_NONE;
}

#define EXP_VELOCITY  100
#define EXP_JUMP    150
/*
===================
CG_BigExplode

Generated a bunch of gibs launching out from the bodies location
===================
*/
void CG_BigExplode( vec3_t playerOrigin )
{
  vec3_t  origin, velocity;

  if( !cg_blood.integer )
    return;

  VectorCopy( playerOrigin, origin );
  velocity[ 0 ] = crandom( ) * EXP_VELOCITY;
  velocity[ 1 ] = crandom( ) * EXP_VELOCITY;
  velocity[ 2 ] = EXP_JUMP + crandom( ) * EXP_VELOCITY;
  CG_LaunchExplode( origin, velocity, cgs.media.smoke2 );

  VectorCopy( playerOrigin, origin );
  velocity[ 0 ] = crandom( ) * EXP_VELOCITY;
  velocity[ 1 ] = crandom( ) * EXP_VELOCITY;
  velocity[ 2 ] = EXP_JUMP + crandom( ) * EXP_VELOCITY;
  CG_LaunchExplode( origin, velocity, cgs.media.smoke2 );

  VectorCopy( playerOrigin, origin );
  velocity[ 0 ] = crandom( ) * EXP_VELOCITY * 1.5;
  velocity[ 1 ] = crandom( ) * EXP_VELOCITY * 1.5;
  velocity[ 2 ] = EXP_JUMP + crandom( ) * EXP_VELOCITY;
  CG_LaunchExplode( origin, velocity, cgs.media.smoke2 );

  VectorCopy( playerOrigin, origin );
  velocity[ 0 ] = crandom( ) * EXP_VELOCITY * 2.0;
  velocity[ 1 ] = crandom( ) * EXP_VELOCITY * 2.0;
  velocity[ 2 ] = EXP_JUMP + crandom( ) * EXP_VELOCITY;
  CG_LaunchExplode( origin, velocity, cgs.media.smoke2 );

  VectorCopy( playerOrigin, origin );
  velocity[ 0 ] = crandom( ) * EXP_VELOCITY * 2.5;
  velocity[ 1 ] = crandom( ) * EXP_VELOCITY * 2.5;
  velocity[ 2 ] = EXP_JUMP + crandom( ) * EXP_VELOCITY;
  CG_LaunchExplode( origin, velocity, cgs.media.smoke2 );
}


//TA: wolf source rip

// just a bunch of numbers we can use for pseudo-randomizing based on time
#define  NUMRANDTABLE 257
unsigned int randtable[ NUMRANDTABLE ] =
{
  0x0000,  0x1021,  0x2042,  0x3063,  0x4084,  0x50a5,  0x60c6,  0x70e7,
  0x8108,  0x9129,  0xa14a,  0xb16b,  0xc18c,  0xd1ad,  0xe1ce,  0xf1ef,
  0x1231,  0x0210,  0x3273,  0x2252,  0x52b5,  0x4294,  0x72f7,  0x62d6,
  0x9339,  0x8318,  0xb37b,  0xa35a,  0xd3bd,  0xc39c,  0xf3ff,  0xe3de,
  0x2462,  0x3443,  0x0420,  0x1401,  0x64e6,  0x74c7,  0x44a4,  0x5485,
  0xa56a,  0xb54b,  0x8528,  0x9509,  0xe5ee,  0xf5cf,  0xc5ac,  0xd58d,
  0x3653,  0x2672,  0x1611,  0x0630,  0x76d7,  0x66f6,  0x5695,  0x46b4,
  0xb75b,  0xa77a,  0x9719,  0x8738,  0xf7df,  0xe7fe,  0xd79d,  0xc7bc,
  0x48c4,  0x58e5,  0x6886,  0x78a7,  0x0840,  0x1861,  0x2802,  0x3823,
  0xc9cc,  0xd9ed,  0xe98e,  0xf9af,  0x8948,  0x9969,  0xa90a,  0xb92b,
  0x5af5,  0x4ad4,  0x7ab7,  0x6a96,  0x1a71,  0x0a50,  0x3a33,  0x2a12,
  0xdbfd,  0xcbdc,  0xfbbf,  0xeb9e,  0x9b79,  0x8b58,  0xbb3b,  0xab1a,
  0x6ca6,  0x7c87,  0x4ce4,  0x5cc5,  0x2c22,  0x3c03,  0x0c60,  0x1c41,
  0xedae,  0xfd8f,  0xcdec,  0xddcd,  0xad2a,  0xbd0b,  0x8d68,  0x9d49,
  0x7e97,  0x6eb6,  0x5ed5,  0x4ef4,  0x3e13,  0x2e32,  0x1e51,  0x0e70,
  0xff9f,  0xefbe,  0xdfdd,  0xcffc,  0xbf1b,  0xaf3a,  0x9f59,  0x8f78,
  0x9188,  0x81a9,  0xb1ca,  0xa1eb,  0xd10c,  0xc12d,  0xf14e,  0xe16f,
  0x1080,  0x00a1,  0x30c2,  0x20e3,  0x5004,  0x4025,  0x7046,  0x6067,
  0x83b9,  0x9398,  0xa3fb,  0xb3da,  0xc33d,  0xd31c,  0xe37f,  0xf35e,
  0x02b1,  0x1290,  0x22f3,  0x32d2,  0x4235,  0x5214,  0x6277,  0x7256,
  0xb5ea,  0xa5cb,  0x95a8,  0x8589,  0xf56e,  0xe54f,  0xd52c,  0xc50d,
  0x34e2,  0x24c3,  0x14a0,  0x0481,  0x7466,  0x6447,  0x5424,  0x4405,
  0xa7db,  0xb7fa,  0x8799,  0x97b8,  0xe75f,  0xf77e,  0xc71d,  0xd73c,
  0x26d3,  0x36f2,  0x0691,  0x16b0,  0x6657,  0x7676,  0x4615,  0x5634,
  0xd94c,  0xc96d,  0xf90e,  0xe92f,  0x99c8,  0x89e9,  0xb98a,  0xa9ab,
  0x5844,  0x4865,  0x7806,  0x6827,  0x18c0,  0x08e1,  0x3882,  0x28a3,
  0xcb7d,  0xdb5c,  0xeb3f,  0xfb1e,  0x8bf9,  0x9bd8,  0xabbb,  0xbb9a,
  0x4a75,  0x5a54,  0x6a37,  0x7a16,  0x0af1,  0x1ad0,  0x2ab3,  0x3a92,
  0xfd2e,  0xed0f,  0xdd6c,  0xcd4d,  0xbdaa,  0xad8b,  0x9de8,  0x8dc9,
  0x7c26,  0x6c07,  0x5c64,  0x4c45,  0x3ca2,  0x2c83,  0x1ce0,  0x0cc1,
  0xef1f,  0xff3e,  0xcf5d,  0xdf7c,  0xaf9b,  0xbfba,  0x8fd9,  0x9ff8,
  0x6e17,  0x7e36,  0x4e55,  0x5e74,  0x2e93,  0x3eb2,  0x0ed1,  0x1ef0
};

#define  LT_MS    100  // random number will change every LT_MS millseconds
#define  LT_RANDMAX  ((unsigned short)0xffff)

float lt_random( int thisrandseed, int t )
{
  return (float)randtable[ abs( ( thisrandseed + t + ( cg.time / LT_MS ) *
                                ( cg.time / LT_MS ) ) ) % NUMRANDTABLE ] / (float)LT_RANDMAX;
}

float lt_crandom( int thisrandseed, int t )
{
  return ( ( 2.0 * ( (float)randtable[ abs( ( thisrandseed + t + ( cg.time / LT_MS ) *
                            ( cg.time / LT_MS ) ) ) % NUMRANDTABLE ] / (float)LT_RANDMAX ) ) - 1.0 );
}

/*
===============
CG_DynamicLightningBolt
===============
*/
void CG_DynamicLightningBolt( qhandle_t shader, vec3_t start, vec3_t pend,
                              int numBolts, float maxWidth, qboolean fade,
                              float startAlpha, int recursion, int randseed )
{
  int i,j;
  float  segMin, segMax, length;
  float  thisSeg, distLeft, thisWidth;
  vec3_t  pos, vec, end;
  int    curJunc;
  vec3_t c, bc = {0.8,0.9,1};
  const  float  rndSize = 12.0;
  const  float  maxSTscale = 30.0;
  float  stScale;
  float  alpha, viewDist;
  const int  trailLife = 1;
  int    forks=0;
  #define  STYPE_LIGHTNING    STYPE_REPEAT  // ST mapping for trail
  #define  FORK_CHANCE    0.5
  #define  VIEW_SCALE_DIST  128

  VectorCopy( pend, end );  // need this so recursive calls don't override stacked endpoints

  length = Distance( start, end );
  //if (length > 128) {
    segMin = length / 10.0;

    if( segMin < 8 )
      segMin = 8;

    segMax = segMin * 1.2;
    //segMax = length / 30.0;
  //} else {
  //  segMin = length / 3.0;
  //  segMax = length / 1.5;
  //}

  if( startAlpha > 1.0 )
    startAlpha = 1.0;

  alpha = startAlpha;  // change only if fading

  for( i = 0; i < numBolts; i++ )
  {
    distLeft = length;
    VectorCopy( start, pos );
    // drop a start junction
    stScale = maxSTscale * ( 0.5 + lt_random( randseed, i + 1 ) * 0.5 );

    if( fade )
    {
      if( startAlpha == 1.0 )
      {
        alpha = startAlpha * ( distLeft / length );
      }
      else
      {
        alpha = 1.0 - 1.0 * fabs( ( 1.0 - ( distLeft / length ) ) - startAlpha );
        if( alpha < 0 )
          alpha = 0;

        if( alpha > 1 )
          alpha = 1;
      }
    }

    thisWidth = maxWidth * ( 0.5 + 0.5 * alpha );

    if( ( viewDist = VectorDistance( pos, cg.refdef.vieworg ) ) < VIEW_SCALE_DIST )
    {
      thisWidth *= 0.5 + ( 0.5 * ( viewDist / VIEW_SCALE_DIST ) );

      if( thisWidth < 4.0 && thisWidth < maxWidth )
        thisWidth = 4.0;
    }
    else
    {  // scale it wider with distance so it remains visible
      thisWidth *= 0.5 + ( 0.5 * ( viewDist / VIEW_SCALE_DIST ) );

      if( thisWidth > maxWidth * 2 )
        // thisWidth > maxWidth*2;
        thisWidth = maxWidth * 2;
    }

    //
    VectorScale( bc, alpha, c );
    c[ 2 ] *= 1.0 + ( 1.0 - alpha );
    if( c[ 2 ] > 1.0 )
      c[ 2 ] = 1.0;
    //
    curJunc = CG_AddTrailJunc( 0, shader, cg.time, STYPE_LIGHTNING, pos, trailLife, 1, 1,
                               thisWidth, thisWidth, TJFL_NOCULL, c, c, stScale, 20.0 );

    while( distLeft > 0 )
    {  // create this bolt
      thisSeg = segMin + ( segMax - segMin ) * lt_random( randseed, 2 );
      thisWidth = maxWidth * ( 0.5 + 0.5 * alpha );

      if( thisSeg >= distLeft - rndSize )
      {  // go directly to the end point
        VectorCopy( end, pos );
        thisWidth *= 0.6;
      }
      else if( (float)distLeft / length < 0.3 )
      {
        VectorSubtract( end, pos, vec );
        VectorNormalize( vec );
        VectorMA( pos, thisSeg, vec, pos );
        // randomize the position a bit
        thisSeg *= 0.05;
        if( thisSeg > rndSize )
          thisSeg = rndSize;

        for( j = 0; j < 3; j++ )
        {
          viewDist = lt_crandom( randseed * randseed, j * j + i * i + 3 );

          if( fabs( viewDist ) < 0.5 )
          {
            if( viewDist > 0 )
              viewDist = 0.5;
            else
              viewDist = -0.5;
          }

          pos[ j ] += viewDist * thisSeg;
        }
      }
      else
      {
        VectorSubtract( end, pos, vec );
        VectorNormalize( vec );
        VectorMA( pos, thisSeg, vec, pos );
        // randomize the position a bit
        thisSeg *= 0.25;

        if( thisSeg > rndSize )
          thisSeg = rndSize;

        for( j = 0; j < 3; j++ )
        {
          viewDist = lt_crandom( randseed, j * j + i * i + 3 );

          if( fabs( viewDist ) < 0.5 )
          {
            if( viewDist > 0 )
              viewDist = 0.5;
            else
              viewDist = -0.5;
          }

          pos[ j ] += viewDist * thisSeg;
        }
      }

      distLeft = Distance( pos, end );

      if( fade )
      {
        if( startAlpha == 1.0 )
        {
          alpha = startAlpha * ( distLeft / length );
        }
        else
        {
          alpha = 1.0 - 1.0 * fabs( ( 1.0 - ( distLeft / length ) ) - startAlpha );
          if( alpha < 0 )
            alpha = 0;

          if( alpha > 1 )
            alpha = 1;
        }
      }

      //thisWidth *= alpha;
      if( ( viewDist = VectorDistance( pos, cg.refdef.vieworg ) ) < VIEW_SCALE_DIST )
      {
        thisWidth *= 0.5 + ( 0.5 * ( viewDist / VIEW_SCALE_DIST ) );

        if( thisWidth < 4.0 && thisWidth < maxWidth )
          thisWidth = 4.0;
      }
      else
      {  // scale it wider with distance so it remains visible
        thisWidth *= 0.5 + ( 0.5 * ( viewDist / VIEW_SCALE_DIST ) );

        if( thisWidth > maxWidth * 2 )
          // thisWidth > maxWidth*2;
          thisWidth = maxWidth * 2;
      }

      //
      VectorScale( bc, alpha, c );
      c[ 2 ] *= 1.0 + ( 1.0 - alpha );

      if( c[ 2 ] > 1.0 )
        c[ 2 ] = 1.0;
      //
      //stScale = maxSTscale * (0.4 + lt_random(randseed,1)*0.6);
      curJunc = CG_AddTrailJunc( curJunc, shader, cg.time, STYPE_LIGHTNING, pos, trailLife,
                                 1, 1, thisWidth, thisWidth, TJFL_NOCULL, c, c, stScale, 20.0 );

      // fork from here?
      if( thisWidth < 4 && distLeft > 10 && recursion < 3 && forks < 3 &&
          lt_random( randseed, 383 + i + forks ) < FORK_CHANCE )
      {
        vec3_t  fend;

        forks++;
        VectorSet( fend,
                   distLeft * 0.3 * lt_crandom( randseed, 56 + i + forks ),
                   distLeft * 0.3 * lt_crandom( randseed, 160 + i + forks ),
                   distLeft * 0.3 * lt_crandom( randseed, 190 + i + forks ) );
        VectorAdd( fend, end, fend );
        VectorSubtract( fend, pos, fend );
        VectorMA( pos, 0.2 + 0.7 * lt_random( randseed, 6 + i + forks ), fend, fend );

        //if (recursion > 0 && recursion < 2) {
        //  CG_DynamicLightningBolt( cgs.media.lightningBoltShader, pos, fend, 1, maxWidth, qtrue, alpha, recursion, randseed );
        //  return;  // divert bolt rather than split
        //} else {
          CG_DynamicLightningBolt( shader, pos, fend, 1, maxWidth, qtrue, alpha,
                                   recursion + 1, randseed + 765 );
        //}
      }

      randseed++;
    }
  }
}
